Attribute VB_Name = "SeaTalk"
Option Explicit
''Module to monitor RayMarine SeaTalk protocol
''
''1. Add this module to your code
''
''2. Make sure you have a CommunicationsControl on your form
''     (for this example MSComm2)
''
''3. Include the following code in your form
''
''3a.:
''
''Private Sub MSComm2_OnComm()
''Dim m As String, n as integer
''
''    MSCommSeaTalkChar MSComm2, SeaTalkModeBoat,m,n
''
''    process m as desired, for example
''    if m <> "" then msgbox m
''End Sub
''
''3b. (On load form Open the desired Com Port using SetSeaTalkCom):
''
''SetSeaTalkCom 1, MSComm2
''
'' Ready to go !!!
''
''The MSCommSeaTalk_OnComm event will monitor the port,
''and calls MSCommSeaTalkChar whenever characters are received
''
''When a message is complete, SCommSeaTalkChar calls DoSeaTalkMessage,
''which will return legible string
''
''
''see Robert Knauf http://www.thomasknauf.de/seatalk.htm
''Philip Beekman, April 2003, p.beekman@chello.nl
Public Const SeaTalkModeBoat = 1
Public Const SeaTalkModeSim = 2

Type ShipData
    LAT As Double  'latitude in DD.DDD  N+/S-
    LON As Double  'longitude in DD.DDD E+/W-
    SOG As Single  'speed Over Ground (knots)
    COG As Single  'Course Over Ground 0-360 degrees
    HDG As Single  'Heading 0-360 degrees
    GMT As Date    'GMT Hour + Date
    AWA As Single  'Apparent Wind Angle
    AWS As Single  'Apparent wind speed (knots)
    TWA As Single  'True Wind Angle
    TWS As Single  'True Wind Speed
    TWD As Single  'Geo Wind Dir
    VW As Single   'Speed Through Water (knots)
    Trip As Single 'Trip Distance (nautical miles)
    TLog As Single 'Total Log (nautical miles)
    DBT As Single  'Depth Below Transducer in meters
    WT As Single   'Water Temperature (Celcius)
    
    BTW As Single  'Bearing to Waypoint (from current position?)
    XTE As Single  'Cross Track Error NM
    DTW As Single  'Distance to Waypoint NM
    WPN As String  'Target Waypoint Name
    
End Type

Public myShipData As ShipData 'will contain last value received, for processing in other functions

Const MetersPerFoot = 0.3
Const pi = 3.14159265358979

Dim msgShown As Boolean

Public Function SetSeaTalkCom(myPort As Integer, myCom As MSComm) As Boolean
'opens myPort for communicationscontrol myCom with settings appropriate to Seatalk
With myCom
On Error GoTo openerr
    If Not msgShown Then
        MsgBox "SeaTalk option may not work well if other software, " & _
        "esp. hyperterminal, has used COM port before; " & _
        "you may want to restart your computer before running in this mode"
        msgShown = True
    End If
    If .PortOpen Then .PortOpen = False
    .CommPort = myPort
    
    .Settings = "4800,S,8,1"
    ' 4800 baud, Space, 8 data, and 1 stop bit.
    'seems to be 4800 baud for SeaTalk according to Knauf, although RayMarine says 9600
    'no Odd/Even parity checking, but check if bit is set to detect start of datagram
    '8 bit data
    '1 stop bit
    
    .ParityReplace = ""
    'Parity will be detected by OnComm event, character will still be used

    .InputMode = comInputModeBinary
    'will treat data as strictly binary

    .Handshaking = comNone
    'SeaTalk has no handshaking
    
    .InputLen = 1
    'We will process 0ne character at a time
    .RThreshold = 1
    'We will react when we get the character
    
    .PortOpen = True
    .Output = ""
    On Error GoTo 0
    SetSeaTalkCom = True
    Exit Function
openerr:
    MsgBox Err.Description & vbCrLf & "COM" & myPort & " : " & .Settings, , "error opening port "
End With
End Function

Public Sub MSCommSeaTalkChar(myComm As MSComm, SeaTalkMode As Integer, mymsg As String, myMsgNum As Integer, Optional myInp As String)
    Static RecBuff(255) As Byte 'place to store characters
    Static nRecChar As Integer 'number of characters received in current datagram
    Static nTotChar As Integer 'total number of characters expected in this datagram
    Static blnParity As Boolean
    Static nUnrecognised As Integer
    Dim blnSignal As Boolean
    Dim btInputByte As Byte
    Dim sc As String
    'these fields only needed to assist in output processing
    
    Dim i As Integer
    mymsg = ""
    With myComm
        Select Case .CommEvent
        Case comEventRxParity:
            blnParity = True 'store event for next call
        Case comEvReceive: 'there must be at least one byte to receive
            btInputByte = .Input(0)
            sc = " "
            'first lets see if we need to recognise a new datagram, depending on SeaTalkMode
            Select Case SeaTalkMode
            Case SeaTalkModeBoat:
                    'note that we don't use the parity bit in its proper form. It is set
                    'to signal the start of a new datagram.
                    'so if number of databits is even, then parity bit is 0 if no new datagram
                    'and if number of databits is odd, then parity bit is 0 means new datagram
                    If blnParity Then
                        If NbitsSetEven(btInputByte) Then
                            blnSignal = True: sc = "P" 'parity was signalled on even bitcount
                        Else
                            ''even here we seem to get a new datagram. Why is not clear
                            blnSignal = True: sc = "Q" 'parity was signalled on odd bitcount
                        End If
                    Else
                        If NbitsSetEven(btInputByte) Then
                            blnSignal = False: sc = "r" 'no parity was signaled on even bitcount
                        Else
                            blnSignal = True: sc = "S" 'no parity was signalled on odd bitcount
                        End If
                    End If
            Case SeaTalkModeSim:
                    'but in case of SeaSigma utility, this does not work this way.
                    'it simply reverses the parity on the first byte
                    blnSignal = blnParity: sc = IIf(blnParity, "f", " ")
                    'but this is if we think we only need to check for parity errors
            End Select 'case SeaTalkMode
            ''pick up first character in buffer
            myInp = myInp & sHexStr(btInputByte, sc)
            blnParity = False 'clear parity bit for future calls
            
            ''If a new datagram was signalled, then prepare for it:
            If blnSignal Then
                'this is start of new datagram
                If nRecChar > 0 Then
                    'we have unfinished datagram left, but don't want to do any processing
                End If
                nRecChar = 0
                nTotChar = 255 'meaning: 255=unknown
            End If
            
            'we will process the first and any remaining other characters in buffer at this stage.
            'in principle there should only be one character.
            'if we have more than one waiting, our computer may be to slow
            'in any case we do not assume additional parity checks needed
            'and just assume they all go with the message.
            Do
                If nRecChar <= 255 Then
                    RecBuff(nRecChar) = btInputByte
                        nRecChar = nRecChar + 1
                    If nRecChar = 2 Then
                        'Total Number of characters = Lower Half of Second byte + 3
                        nTotChar = (RecBuff(1) And &HF) + 3
                    End If
                    If nTotChar = nRecChar Then
                        ''so we now have all the characters of the SeaTalk message
                        If nUnrecognised > 0 Then
                            'we have unrecognised characters, but don't process these
                            nUnrecognised = 0
                        End If
                        If mymsg <> "" Then mymsg = mymsg & vbCrLf
                        mymsg = mymsg & PadStr(HexStr(RecBuff, nTotChar), 30) & _
                            "m:" & DoSeaTalkMessage(RecBuff)
                        myMsgNum = RecBuff(0)
                        nRecChar = 0
                        nTotChar = -1
                    Else
                        If nRecChar > nTotChar Then
                            ''input unrecognised characters after Seatalk DataGram : " & nRecChar & " " & nTotChar
                            nUnrecognised = nUnrecognised + 1
                            nTotChar = -1
                            nRecChar = 0
                        Else
                            'nothing wrong here, regular character, but not yet all received
                        End If
                    End If
                Else
                    'inputbuffer overrun, discard
                    nRecChar = 0
                End If
                If .InBufferCount > 0 Then
                    ''pick up remaining character in buffer
                    btInputByte = .Input(0)
                    myInp = myInp & sHexStr(btInputByte, "x")
                Else
                    ''nothing left in buffer, then return
                    Exit Do
                End If
            Loop
        End Select 'Case .CommEvent
    End With
End Sub



Function DoSeaTalkMessage(b() As Byte) As String
'assumes as correct message is received in B, and interprets its value
Dim strResult As String

Select Case (b(0))
''selected NMEA sentences
Case &H0:
    ''00  02  YZ  XX  XX  Depth below transducer: XXXX/10 feet
    ''                     Display units: Y=0 => feet, Y=4 => meter
    ''                     Flags: Z&1 Shalow Depth Alarm   (Z=1)
    ''                            Z&2 Transducer defective (Z=4)
    ''                     Corresponding NMEA sentences: DPT, DBT
    myShipData.DBT = ((b(4) * 256# + b(3)) / 10) * MetersPerFoot
    strResult = "DBT 00: " & myShipData.DBT & " feet " & b(1)
Case &H10:
    ''10  01  XX  YY  Apparent Wind Angle: XXYY/2 degrees right of bow
    ''                 Used for autopilots Vane Mode (WindTrim)
    ''                 Corresponding NMEA sentence: MWV
    myShipData.AWA = (b(2) * 256# + b(3)) / 2
    strResult = "AWA 10: " & myShipData.AWA & " degrees right of bow"
Case &H11:
    '' 11  01  XX  0Y  Apparent Wind Speed: (XX & 0x7F) + Y/10 Knots
    ''                 Units flag: XX&0x80=0    => Display value in Knots
    ''                             XX&0x80=0x80 => Display value in Meter/Second
    ''                 Corresponding NMEA sentence: MWV
    myShipData.AWS = (b(2) Mod 128) + b(3) / 10
    strResult = "AWS 11: " & myShipData.AWS & " knots, display " & _
                    IIf((b(2) And &H80) = 0, "knots", "meters/second")
Case &H20:
    '' 20  01  XX  XX  Speed through water: XXXX/10 Knots
    ''                 Corresponding NMEA sentence: VHW
    myShipData.VW = (b(3) * 256# + b(2)) / 10
    strResult = "WS  20: " & myShipData.VW & " knots"
Case &H21:
    '' 21  02  XX  XX  0X  Trip Mileage: XXXXX/100 nautical miles
    myShipData.Trip = ((b(4) * 256# + b(3)) * 256# + b(2)) / 100
    strResult = "Trip21: " & myShipData.Trip & " NM"
Case &H22:
    myShipData.TLog = ((b(4) * 256# + b(3)) * 256# + b(2)) / 10
    '' 22  02  XX  XX  00  Total Mileage: XXXX/10 nautical miles
    strResult = "TLOG20: " & myShipData.TLog & " NM"
Case &H23:
''23  41  XX  YY  Water temperature (ST50): XX deg Celsius, YY deg Fahrenheit
''                 Corresponding NMEA sentence: MTW
    strResult = "Temp : " & b(2) & " C, " & b(3) & " F"
Case &H24:
'' 24  02  00  00  XX  Display units for Mileage & Speed
''                     XX: 00=nm/knots, 06=sm/mph, 86=km/kmh
    If b(4) = &H0 Then strResult = "Display Units = nm/knots"
    If b(4) = &H6 Then strResult = "Display Units = sm/mph"
    If b(4) = &H86 Then strResult = "Display Units = km/kmh"
Case &H25:
     ''25  Z4  XX  YY  UU  VV AW  Total & Trip Log
     ''                    total= (XX+YY*256+Z* 4096)/ 10 [max=104857.5] nautical miles
     ''                    trip = (UU+VV*256+W*65536)/100 [max=10485.75] nautical miles
    myShipData.TLog = (b(2) + b(3) * 256# + (b(1) \ 16) * 4096) / 10
    myShipData.Trip = (b(4) + b(5) * 256# + b(6) * 65536) / 100
    strResult = "Total & Trip 25: " & myShipData.TLog & "   " & _
                            myShipData.Trip & " NM"
Case &H26:
    ''26  04  XX  XX  YY  YY D1  Speed through water:
    ''                    Sensor 1: XXXX/100 Knots, valid if D1&64=64
    ''                    Sensor 2: YYYY/100 Knots, valid if D1&128=128
    ''                    Corresponding NMEA sentence: VHW
    strResult = "WS  26:" & IIf((b(7) And 64) = 64, (b(3) * 256# + b(2)) / 100, "N/A") & " " & IIf((b(7) And 128) = 128, (b(5) * 256# + b(4)) / 100, "N/A") & " knots"
Case &H27:
    ''27  01  XX  XX  Water temperature: (XXXX-100)/10 deg Celsius
    ''                 Corresponding NMEA sentence: MTW
    myShipData.WT = (b(2) + b(3) * 256# - 100) / 10
    strResult = "MTW 27: " & myShipData.WT & " deg Celsius"
Case &H50:
    ''50  A2  XX  YY  YY  LAT position: XX degrees, (YYYY & 0x7FFF)/100 minutes
    ''                    MSB of Y = YYYY & 0x8000 = South if set, North if cleared
    ''                     Corresponding NMEA sentences: RMC, GAA, GLL
    myShipData.LAT = b(2) + (((b(4) Mod 128) * 256# + b(3)) / 6000)
    If b(4) And &H80 Then myShipData.LAT = -myShipData.LAT
    strResult = "LAT: " & b(2) & IIf(b(4) And &H80, "S", "N") & ((b(4) Mod 128) * 256# + b(3)) / 100 & "   "
Case &H51:
    ''51  A2  XX  YY  YY  LON position: XX degrees, (YYYY & 0x7FFF)/100 minutes
    ''                              MSB of Y = YYYY & 0x8000 = East if set, West if cleared
    ''                 Corresponding NMEA sentences: RMC, GAA, GLL
    myShipData.LON = b(2) + (((b(4) Mod 128) * 256# + b(3)) / 6000)
    If Not (b(4) And &H80) Then myShipData.LON = -myShipData.LON
    strResult = "LON: " & b(2) & IIf(b(4) And &H80, "E", "W") & ((b(4) Mod 128) * 256# + b(3)) / 100 & "   "
Case &H52:
    ''52  01  XX  XX  Speed over Ground: XXXX/10 Knots
    ''                Corresponding NMEA sentences: RMC, VTG
    myShipData.SOG = (b(3) * 256# + b(2)) / 10
    strResult = "SOG 52: " & myShipData.SOG & " knots"
Case &H53:
    '' 53  X0  XX      Course Magnetic: XXX/8 Degrees
    ''                 Least significant 2 bits are always 0,
    ''                 giving a resolution of 0.5 degrees
    ''                Corresponding NMEA sentences: RMC, VTG
    ''PRB20030512  this is the most likely formula I can come up with:
    myShipData.COG = (b(2) * 8#) + (b(1) \ 16) / 2
    strResult = "COG 53: " & myShipData.COG & " degrees - check if this is really COG"
Case &H54:
    ''54  T1  TT  HH  GMT-time: HH hours,
    ''                          6 MSBits of TTT = (TTT & 0xFC0) / 64 = minutes
    ''                           6 LSBits of TTT = TTT & 0x3F = seconds
    ''                 Corresponding NMEA sentences: RMC, GAA, BWR, BWC
    
    myShipData.GMT = Int(myShipData.GMT) + CDate(b(3) & ":" & b(2) \ 4 & ":" & (b(2) And &H3) * 16# + b(1) \ 16)
    strResult = "GMT :" & Format(myShipData.GMT, "HH:MM:SS") & " HH:MM:SS"
''Case &H55:
'' 55  X1  YY  yy  TRACK keystroke on GPS unit
''                 keycodes identical with autopilot (command 86)
Case &H56:
    '' 56  M1  DD  YY  Date: YY year, M month, DD day in month
    ''                 Corresponding NMEA sentence: RMC
    myShipData.GMT = (myShipData.GMT - Int(myShipData.GMT)) + DateSerial(2000 + b(3), b(1) \ 16, b(2))
    strResult = "GMT :" & Format(myShipData.GMT, "YYYYMMDD") & "  YYYYMMDD"
    ''strResult = "GMT :" & "20" & Format(B(3), "00") & Format(B(1) \ 16, "00") & Format(B(2), "00") & "  YYYYMMDD"
''Case &H57:
'' 57  S0  DD      Sat Info: S number of sats, DD horiz. dillution of position
''                 Corresponding NMEA sentences: GGA, GSA
Case &H58:
    '' 58  Z5  LA XX YY LO QQ RR   LAT/LON
    ''                 LA Degrees LAT, LO Degrees LON
    ''                 minutes LAT = (XX * 256# + YY) / 1000
    ''                 minutes LON = (QQ * 256# + RR) / 1000
    ''                 Z&1: South (Z&1 = 0: North)
     ''                Z&2: East  (Z&2 = 0: West)
    ''                 Corresponding NMEA sentences: RMC, GAA, GLL
    
    With myShipData
        .LAT = b(2) + (b(3) * 256# + b(4)) / 60000
        If b(1) And &H10 = &H10 Then .LAT = -.LAT
        .LON = b(5) + (b(6) * 256# + b(7)) / 60000
        If Not (b(1) And &H20 = &H20) Then .LON = -.LON
    End With
    strResult = "POS: " & b(2) & IIf(b(1) And &H10, "S", "N") & (b(3) * 256# + b(4)) / 1000 & "   " & _
                          b(5) & IIf(b(1) And &H20, "E", "W") & (b(6) * 256# + b(7)) / 1000

Case &H82
''82  05  XX  xx YY yy ZZ zz   Target waypoint name
''                 XX+xx = YY+yy = ZZ+zz = FF (allows error detection)
''                 Takes the last 4 chars of name, assumes upper case only
''                 Char= ASCII-Char - 0x30
''                 XX&0x3F: char1
''                 (YY&0xF)*4+(XX&0xC0)/64: char2
''                 (ZZ&0x3)*16+(YY&0xF0)/16: char3
''                 (ZZ&0xFC)/4: char4
''                 Corresponding NMEA sentences: RMB, APB, BWR, BWC
If (b(2) + b(3) = &HFF) And (b(4) + b(5) = &HFF) And (b(6) + b(7) = &HFF) Then
myShipData.WPN = Chr((b(2) And &H3F) + &H30) & _
                 Chr((b(4) And &HF) * 4 + (b(2) And &HC0) / 64 + &H30) & _
                 Chr((b(6) And &H3) * 16 + (b(4) And &HF0) / 16 + &H30) & _
                 Chr((b(6) And &HFC) / 4 + &H30)
strResult = "WPN 82: " & myShipData.WPN
End If

Case &H84
'' 84  U6  VW  XY 0Z 00 RR SS TT  Compass heading  Autopilot course and
''                 Rudder position (see also command 9C)
''                  Compass heading in degrees:
''                    The two lower  bits of  U * 90 +
''                    the six lower  bits of VW *  2 +
''                    the two higher bits of  U /  2 =
''                    (U & 0x3) * 90 + (VW & 0x3F) * 2 + (U & 0xC) / 8
''                  Autopilot course in degrees:
''                    The two higher bits of  V * 90 + XY / 2
''                  Z & 0x2 = 0 : Autopilot in Standby-Mode
''                  Z & 0x2 = 2 : Autopilot in Auto-Mode
''                  Z & 0x4 = 4 : Autopilot in Vane Mode (WindTrim), requires regular "10" datagrams
''                  Rudder position: RR degrees (positive values steer right,
''                    negative values steer left. Example: 0xFE = 2 left)
''                  SS & 0x01 : when set, turns off heading display on 600R control.
''                  SS & 0x02 : always on with 400G
''                  SS & 0x08 : displays NO DATA on 600R
''                  SS & 0x10 : displays LARGE XTE on 600R
''                  SS & 0x80 : Displays Auto Rel on 600R
''                  TT : Always 0x08 on 400G computer

myShipData.HDG = ((b(1) \ 16) And &H3) * 90 + (b(2) And &H3F) * 2 + ((b(1) \ 16) And &HC) / 8
strResult = "HDG 84: " & myShipData.HDG & " degrees - check if this is really HDG"

Case &H85
'' 85  X6  XX  VU ZW ZZ YF 00 yf   Navigation to waypoint information
''                  Cross Track Error: XXX/100 nautical miles
''                   Example: X-track error 2.61nm => 261 dec => 0x105 => X6XX=5_10
''                  Bearing to destination: (U & 0x3) * 90 + WV / 2
''                   Example: GPS course 230=180+50=2*90 + 0x64/2 => VUZW=42_6
''                   U&8: U&8 = 8 -> Bearing is true, U&8 = 0 -> Bearing is magnetic
''                  Distance to destination: Distance 0-9.99nm: ZZZ/100nm, Y & 1 = 1
''                                           Distance >=10.0nm: ZZZ/10 nm, Y & 1 = 0
''                  Direction to steer: if Y & 4 = 4 Steer right to correct error
''                                      if Y & 4 = 0 Steer left  to correct error
''                  Example: Distance = 5.13nm, steer left: 5.13*100 = 513 = 0x201 => ZW ZZ YF=1_ 20 1_
''                           Distance = 51.3nm, steer left: 51.3*10  = 513 = 0x201 => ZW ZZ YF=1_ 20 0_
''                  Track control mode:
''                     F= 0x1: Display x-track error and Autopilot course
''                     F= 0x3: Enter Track Control Mode, i.e. lock on to GPS.
''                             Display x-track error, autopilot course and bearing
''                             to destination
''                     F= 0x5: Display x-track error, distance to waypoint,
''                             autopilot course and bearing to destination
''           normal--> F= 0x7: Enter Track Control Mode, i.e. lock on to GPS.
''                             Display x-track error, distance to waypoint,
''                             autopilot course and bearing to destination
''                     F= 0xF: As 0x7 but with x-track error alarm
''                     F= 2, 4, 6, 8 ... causes data errors
''                   Corresponding NMEA sentences: RMB, APB, BWR, BWC, XTE
myShipData.XTE = ((b(1) \ 16 + 256 * b(2)) / 100) * IIf((b(4) \ 16 And 4) = 4, 1, -1)
myShipData.BTW = (b(3) And &H3) * 90 + ((b(3) \ 16) + (b(4) And &HF) * 16) / 2
myShipData.DTW = (b(4) \ 16 + b(5) * 16) / IIf(b(6) \ 16 And 1, 100, 10)
strResult = "NAV 85: XTE=" & myShipData.XTE & " BTW=" & myShipData.BTW & " DTW=" & myShipData.DTW
Case Else
    ''message ignored / not decoded
    strResult = ""
End Select
With myShipData
    If .SOG <> 0 Then
        ApparantToTrue .AWA, .AWS, .SOG, .COG, .TWA, .TWS, .TWD
    Else
        ApparantToTrue .AWA, .AWS, .VW, .HDG, .TWA, .TWS, .TWD
    End If
End With
DoSeaTalkMessage = strResult
End Function
Sub ApparantToTrue(AWA As Single, AWS As Single, SPD As Single, COG As Single, TWA As Single, TWS As Single, TWD As Single)
'calculates TWA and TWS given the first three parameters
Dim d As Double, r As Double
    r = pi * AWA / 180
    If AWS <> SPD * Cos(r) Then
        d = Atn((SPD * Sin(r)) / (AWS - SPD * Cos(r)))
        TWS = AWS * Cos(d) - SPD * Cos(d + r)
    Else
        TWS = 0
    End If
    TWA = 180 * (d + r) / pi
    TWD = (COG + TWA) Mod 360
End Sub
Sub WriteSeaTalkLog(b() As Byte, N As Integer, msg As String)
    Dim oStr As String, HexStr As String, i As Integer
    oStr = Now & " "
    For i = 0 To N - 1
        HexStr = Hex(b(i)): If Len(HexStr) = 1 Then HexStr = "0" & HexStr
        oStr = oStr & HexStr & " "
    Next i
    oStr = oStr & vbCrLf & String(20, " ") & msg
    Debug.Print oStr
    ''Open myTrackPath & "\" & Format(getUTC, "YYYYMMDD") & ".stt" For Append As #4
    ''Print #4, oStr
    ''Close #4
End Sub

Function NbitsSetEven(ByVal b As Byte) As Boolean
Dim i As Integer, c As Integer
    For i = 0 To 7
        If b And &H1 Then c = c + 1
        b = b \ 2
    Next i
    NbitsSetEven = (c Mod 2) = 0
End Function
Function HexStr(b() As Byte, N As Integer, Optional char As String = " ") As String
    Dim i As Integer, s As String, h As String
    For i = 0 To N - 1
        h = Hex(b(i)): If Len(h) = 1 Then h = "0" & h
        s = s & h & char
    Next i
    HexStr = s
End Function
Function sHexStr(b As Byte, Optional char As String = " ") As String
    Dim i As Integer, s As String, h As String
    h = Hex(b): If Len(h) = 1 Then h = "0" & h
    If char = "P" Then s = s & vbCrLf
    s = s & h & char
    sHexStr = s
End Function
Function PadStr(ByVal s As String, N As Integer) As String
    If Len(s) > N Then
        PadStr = Left(s, N)
    Else
        PadStr = s & Space(N - Len(s))
    End If
End Function



